// RestartViaManager.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

DWORD stringToPid(const wchar_t *szExePath);
// bool pidToPath(std::wstring &strOutExePath, DWORD dwPid);
bool pidToProcStartTime(FILETIME *pft, const DWORD dwPID);
int restartApp(const bool fQuiet, const wchar_t *szExePath, const DWORD dwPID);
int restartApp(const bool fQuiet, const wchar_t *szExePath, RM_UNIQUE_PROCESS *prup, const DWORD dwRMSessionHandle);


int _tmain(int argc, _TCHAR* argv[])
{
	bool fQuiet = false;
	const wchar_t *szExePath = nullptr;

	if (argc == 2)
	{
		szExePath = argv[1];
	}
	else if (argc == 3)
	{
		if (_wcsicmp(argv[1], L"quiet") == 0)
		{
			fQuiet = true;
			szExePath = argv[2];
		}
		else
		if (_wcsicmp(argv[2], L"quiet") == 0)
		{
			fQuiet = true;
			szExePath = argv[1];
		}
	}

	if (szExePath == nullptr)
	{
		wprintf(L"Restarts a process using the Windows Restart Manager API.\n");
		wprintf(L"The process must be running, and must have opted in to being restartable via the Restart Manager.\n");
		wprintf(L"Usage:   RestartViaManager.exe \"<full path to process exe>\" [QUIET]\n");
		wprintf(L"Example: RestartViaManager.exe \"C:\\windows\\explorer.exe\"\n");
		wprintf(L"The optional QUIET argument will mean nothing is printed if the command is successful.\n");
		wprintf(L"You can also give a numeric PID instead of the path to the process.\n");
		wprintf(L"  PID is decimal by default; prefix with 0x for hex.\n");
		wprintf(L"  Example: RestartViaManager.exe 3608\n");
		wprintf(L"  Example: RestartViaManager.exe 0xE18\n");
		return -1;
	}

	if (::CoInitialize(0) != S_OK)
	{
		wprintf(L"RestartViaManager: Failed to initialize COM.\r\n");
		return -1;
	}

	DWORD dwPID = stringToPid(szExePath);
	int iResult = restartApp(fQuiet, szExePath, dwPID);

	::CoUninitialize();

	return iResult;
}

DWORD stringToPid(const wchar_t *szExePath)
{
	bool fHex = false;

	const wchar_t *szNumbers = szExePath;

	if (szNumbers[0] == L'0'
	&&	(szNumbers[1] == L'x' || szNumbers[1] == L'X'))
	{
		fHex = true;
		szNumbers += 2;
	}

	for (const wchar_t *szTestNum = szNumbers; *szTestNum != L'\0'; ++szTestNum)
	{
		if ((fHex && !iswxdigit(*szTestNum))
		||	(!fHex && !iswdigit(*szTestNum)))
		{
			return 0;
		}
	}

	DWORD dwPID = wcstoul(szNumbers, NULL, fHex ? 16: 10);
	if (dwPID == 0 || dwPID == ULONG_MAX)
	{
		return 0;
	}

	return dwPID;
}

/*
bool pidToPath(std::wstring &strOutExePath, const DWORD dwPid)
{
	HANDLE hProc = ::OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwPID);
	if (hProc == NULL)
	{
		DWORD dwErr = ::GetLastError();
		wprintf(L"RestartViaManager: Failed to open PID %ld, error %ld.\r\n", dwPID, dwErr);
		return false;
	}

	bool fResult = false;

	wchar_t szExePath[2048] = {0};
	DWORD dwBufSize = _countof(szExePath);
	if (!::QueryFullProcessImageName(hProc, 0, szExePath, &dwBufSize)
	||	dwBufSize >= _countof(szExePath)
	||	szExePath[0] == L'\0')
	{
		DWORD dwErr = ::GetLastError();
		wprintf(L"RestartViaManager: Failed to query path of PID %ld, error %ld.\r\n", dwPID, dwErr);
	}
	else
	{
		strOutExePath = szExePath;
		fResult = true;
	}

	::CloseHandle(hProc);
	return fResult;
}
*/

bool pidToProcStartTime(FILETIME *pft, const DWORD dwPID)
{
	HANDLE hProc = ::OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwPID);
	if (hProc == NULL)
	{
		DWORD dwErr = ::GetLastError();
		wprintf(L"RestartViaManager: Failed to open PID %ld, error %ld.\r\n", dwPID, dwErr);
		return false;
	}

	bool fResult = false;

	FILETIME creationTime = {0};
	FILETIME exitTime = {0};
	FILETIME kernelTime = {0};
	FILETIME userTime = {0};

	if (!::GetProcessTimes(hProc, &creationTime, &exitTime, &kernelTime, &userTime))
	{
		DWORD dwErr = ::GetLastError();
		wprintf(L"RestartViaManager: Failed to query start time of PID %ld, error %ld.\r\n", dwPID, dwErr);
	}
	else
	{
		*pft = creationTime;
		fResult = true;
	}

	::CloseHandle(hProc);
	return fResult;
}

int restartApp(const bool fQuiet, const wchar_t *szExePath, const DWORD dwPID)
{
	RM_UNIQUE_PROCESS *prup = nullptr;
	RM_UNIQUE_PROCESS rup = {0};

	if (dwPID != 0)
	{
		rup.dwProcessId = dwPID;

		if (!pidToProcStartTime(&rup.ProcessStartTime, dwPID))
		{
			// pidToProcStartTime will have already printed an error message.
			return -1;
		}

		prup = &rup;
	}
	else
	{
		DWORD dwAttr = ::GetFileAttributes(szExePath);
		if (dwAttr == INVALID_FILE_ATTRIBUTES)
		{
			wprintf(L"RestartViaManager: \"%s\" not found.\n", szExePath);
			return -1;
		}

		if (dwAttr & FILE_ATTRIBUTE_DIRECTORY)
		{
			wprintf(L"RestartViaManager: \"%s\" is a directory.\n", szExePath);
			return -1;
		}
	}

	DWORD dwRMSessionHandle = 0;
	wchar_t szRMSessionKey[ max( CCH_RM_SESSION_KEY+1, 1024 ) ] = {0};

	DWORD dwSessionRes = ::RmStartSession(&dwRMSessionHandle, 0, szRMSessionKey);

	if (dwSessionRes != ERROR_SUCCESS)
	{
		wprintf(L"RestartViaManager: Failed to start Restart Manager session. (Error %ld)\n", dwSessionRes);
		return -1;
	}

	int iResult = restartApp(fQuiet, szExePath, prup, dwRMSessionHandle);

	dwSessionRes = ::RmEndSession(dwRMSessionHandle);
	
	if (dwSessionRes != ERROR_SUCCESS) // && iResult == 0)
	{
		wprintf(L"RestartViaManager: Failed to end Restart Manager session. (Error %ld)\n", dwSessionRes);
		return -1;
	}

	return iResult;
}

int restartApp(const bool fQuiet, const wchar_t *szExePath, RM_UNIQUE_PROCESS *prup, const DWORD dwRMSessionHandle)
{
	DWORD dwRes = ERROR_BAD_ARGUMENTS;
	
	if (prup == nullptr && szExePath != nullptr)
	{
		dwRes = ::RmRegisterResources(dwRMSessionHandle, 1, &szExePath, 0, NULL, 0, NULL);
	}
	else if (prup != nullptr)
	{
		dwRes = ::RmRegisterResources(dwRMSessionHandle, 0, NULL, 1, prup, 0, NULL);
	}

	if (dwRes != ERROR_SUCCESS)
	{
		wprintf(L"RestartViaManager: Failed to add \"%s\" to Restart Manager's resource list. (Error %ld)\n", szExePath, dwRes);
		return -1;
	}

	dwRes = ::RmShutdown(dwRMSessionHandle, 0, NULL);

	if (dwRes != ERROR_SUCCESS)
	{
		wprintf(L"RestartViaManager: Failed to shutdown \"%s\" via Restart Manager. (Error %ld)\n", szExePath, dwRes);
		return -1;
	}

	dwRes = ::RmRestart(dwRMSessionHandle, 0, NULL);

	if (dwRes != ERROR_SUCCESS)
	{
		wprintf(L"RestartViaManager: Failed to restart \"%s\" via Restart Manager. (Error %ld)\n", szExePath, dwRes);
		return -1;
	}

	if (!fQuiet)
	{
		wprintf(L"RestartViaManager: Shutdown and restart requests succeeded for \"%s\".\n", szExePath);
	}

	return 0;
}
